'''
This is the Lynxy client package, which allows communication with other Lynxy machines.
It simplifies a lot of the things such as security, ease of communication, and more. 
To find documentation, go to the github~!
- Github: https://github.com/SketchedDoughnut/lynxy
'''

# extending main lynx file
from .lynx import *