# raw string forms
OPERATION_SUCCESS = "000"

OPERATION_FAIL = "001"

END_SESSION = "002"

INVALID_COMMAND = "003"

CRASH = "004"

INVALID_USERNAME_DATA = "100"

INVALID_AUTH_TOKEN = "101"

USER_NOT_AUTHORIZED = "102"

INVALID_MESSAGE = "103"

ALREADY_REGISTERED_USERNAME = "104"

USERNAME_EXISTS = "105"

FEATURE_DISABLED = "106"

KILL_ALL = "200"


# encoded forms
# OPERATION_SUCCESS = OPERATION_SUCCESS.encode()
# OPERATION_FAIL = OPERATION_FAIL.encode()
# END_SESSION = END_SESSION.encode()
# INVALID_COMMAND = INVALID_COMMAND.encode()
# CRASH = CRASH.encode()
# INVALID_USERNAME_DATA = INVALID_USERNAME_DATA.encode()
# INVALID_AUTH_TOKEN = INVALID_AUTH_TOKEN.encode()
# USER_NOT_AUTHORIZED = USER_NOT_AUTHORIZED.encode()
# INVALID_MESSAGE = INVALID_MESSAGE.encode()
# ALREADY_REGISTERED_USERNAME = ALREADY_REGISTERED_USERNAME.encode()
# USERNAME_EXISTS = USERNAME_EXISTS.encode()
# KILL_ALL = KILL_ALL.encode()